--[[
	Copyright (C) Udorn (Blackhand)

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
--]]

vendor.ScannerItemModel = {}
vendor.ScannerItemModel.prototype = {}
vendor.ScannerItemModel.metatable = {__index = vendor.ScannerItemModel.prototype}

local L = vendor.Locale.GetInstance()

local log = vendor.Debug:new("ScannerItemModel")

-- the identifiers of the supported columns
vendor.ScannerItemModel.TEXTURE = 1
vendor.ScannerItemModel.NAME = 2
vendor.ScannerItemModel.COUNT = 3
vendor.ScannerItemModel.OWNER = 4
vendor.ScannerItemModel.BID = 5
vendor.ScannerItemModel.BUYOUT = 6
vendor.ScannerItemModel.REASON = 7

-- the several sort functions
local SORT_FUNCS = {
	[vendor.ScannerItemModel.COUNT] = {
		[true] = function(a, b) 
			return a.count < b.count
		end,
		[false] = function(a, b)
			return a.count > b.count
		end
	},
	[vendor.ScannerItemModel.NAME] = {
		[true] = function(a, b)
			return a.name < b.name
		end,
		[false] = function(a, b)
			return a.name > b.name
		end
	},
	[vendor.ScannerItemModel.OWNER] = {
		[true] = function(a, b)
			return a.owner < b.owner
		end,
		[false] = function(a, b)
			return a.owner > b.owner
		end
	},
	[vendor.ScannerItemModel.BID] = {
		[true] = function(a, b)
			return (math.max(a.minBid, a.bidAmount) / a.count) < (math.max(b.minBid, b.bidAmount) / b.count)
		end,
		[false] = function(a, b)
			return (math.max(a.minBid, a.bidAmount) / a.count) > (math.max(b.minBid, b.bidAmount) / b.count)
		end
	},
	[vendor.ScannerItemModel.BUYOUT] = {
		[true] = function(a, b)
			return (a.buyout / a.count) < (b.buyout / b.count)
		end,
		[false] = function(a, b)
			return (a.buyout / a.count) > (b.buyout / b.count)
		end
	},
	[vendor.ScannerItemModel.REASON] = {
		[true] = function(a, b)
			return a.reason < b.reason
		end,
		[false] = function(a, b)
			return a.reason > b.reason
		end
	}

}

--[[
	Checks whether the list has to be sorted and does it in case.
--]]
local function _EnsureSorting(self)
	if (not self.sorted) then
		table.sort(self.index, self.sortFunc)
		self.sorted = true
		self.updateCount = self.updateCount + 1
	end
end

--[[
	Inits the table of descriptors.
--]]
local function _InitDescriptors(self)
	self.descriptors = {
		[vendor.ScannerItemModel.TEXTURE] = {type = "texture"},
		[vendor.ScannerItemModel.NAME] = {type = "text", align = "LEFT", weight = 35, minWidth = 60, sortable = true, title = L["Name"]},
		[vendor.ScannerItemModel.COUNT] = {type = "number", align = "CENTER", weight = 10, minWidth = 20, sortable = true, title = L["Itms."]},
		[vendor.ScannerItemModel.OWNER] = {type = "text", align = "CENTER", weight = 15, minWidth = 60, sortable = true, title = L["Seller"]},
		[vendor.ScannerItemModel.BID] = {type = "text", align = "RIGHT", weight = 20, minWidth = 70, sortable = true, title = L["Bid"]},
		[vendor.ScannerItemModel.BUYOUT] = {type = "text", align = "RIGHT", weight = 20, minWidth = 70, sortable = true, title = L["Buyout"]},
		[vendor.ScannerItemModel.REASON] = {type = "text", align = "LEFT", weight = 60, minWidth = 100, sortable = true, title = L["Reason"]}
	}
end

--[[
	Creates a new instance.
--]]
function vendor.ScannerItemModel:new()
	local instance = setmetatable({}, self.metatable)
	instance.selectedIndex = {}
	instance.index = {}
	instance.updateCount = 0
	instance.updateListeners = {}
	_InitDescriptors(instance)
	instance:Sort(vendor.ScannerItemModel.BUYOUT, true)
	return instance
end

--[[
	Adds a new item to the model.
--]]
function vendor.ScannerItemModel.prototype:AddItem(itemLink, itemLinkKey, name, texture, timeLeft, count, minBid, minIncrement, buyout, bidAmount, owner, reason)
	log:Debug("AddItem enter reason [%s] owner [%s]", reason, owner)
	table.insert(self.index, {itemLinkKey = itemLinkKey, itemLink = itemLink, texture = texture, name = name, timeLeft = timeLeft, count = count, minBid = minBid, minIncrement = minIncrement, buyout = buyout, bidAmount = bidAmount, owner = owner, reason = reason})
	self.sorted = false
	self.updateCount = self.updateCount + 1
	log:Debug("notify")
	for _,listener in pairs(self.updateListeners) do
    	listener.func(listener.arg)
    end
    log:Debug("exit")
end

--[[
	ItemModel function for returning the desired descriptor.
--]]
function vendor.ScannerItemModel.prototype:GetColumnDescriptor(id)
	return self.descriptors[id]
end

--[[
	Returns the number of items available.
--]]
function vendor.ScannerItemModel.prototype:Size()
	return #self.index
end

--[[
	Returns the value(s) for the given row and column id.
--]]
function vendor.ScannerItemModel.prototype:GetValues(row, id)
	log:Debug("GetValues row [%s] id [%s]", row, id)
	_EnsureSorting(self)
	local info = self.index[row]
	if (id == vendor.ScannerItemModel.TEXTURE) then
		return info.texture, info.itemLink, info.count
	elseif (id == vendor.ScannerItemModel.NAME) then
		return info.name
	elseif (id == vendor.ScannerItemModel.COUNT) then
		return info.count
	elseif (id == vendor.ScannerItemModel.OWNER) then
		return info.owner
	elseif (id == vendor.ScannerItemModel.REASON) then
		return info.reason
	elseif (id == vendor.ScannerItemModel.BID) then
		local bid = math.max(info.minBid, info.bidAmount)
		local msg 
		if (info.count > 1) then
			msg = vendor.Format.FormatMoneyValues(bid / info.count, bid, true)
		else
			msg = vendor.Format.FormatMoney(bid / info.count, true)
		end
		return msg
	elseif (id == vendor.ScannerItemModel.BUYOUT) then
		local msg
		if (info.buyout > 0) then
			if (info.count > 1) then
				msg = vendor.Format.FormatMoneyValues(info.buyout / info.count, info.buyout, true)
			else
				msg = vendor.Format.FormatMoney(info.buyout / info.count, true)
			end
		end
		return msg
	else
		error("unknown id: "..(id or "NA"));
	end
	return nil
end

--[[
	Sorts according to the given column.
--]]
function vendor.ScannerItemModel.prototype:Sort(id, ascending)
	vendor.Vendor:Debug("Sort id: %d ascending: %s", id, ascending)
	local sortable = self.descriptors[id].sortable
	if (sortable) then
		self.sortFunc = SORT_FUNCS[id][ascending]
		self.sorted = false
		self.sortId = id
		self.ascending = ascending
		_EnsureSorting(self)
	end
end

--[[
	Toggles the sort state for the given type. If it's currently not
	the sorting criteria, then it will be activated and set to ascending.
--]]
function vendor.ScannerItemModel.prototype:ToggleSort(id)
	vendor.Vendor:Debug("ToggleSort id: %d", id)
	if (id == self.sortId) then
		self:Sort(id, not self.ascending)
	else
		self:Sort(id, true)
	end
end

--[[
	Returns the sorted id and whether it's ascending.
--]]
function vendor.ScannerItemModel.prototype:GetSortInfo()
	return self.sortId, self.ascending
end

--[[
	Selects the given item.
--]]
function vendor.ScannerItemModel.prototype:SelectItem(row, isSelected)
	_EnsureSorting(self)
	self.index[row].selected = isSelected
	self.updateCount = self.updateCount + 1
end

--[[
	Returns whether the given item was selected.
--]]
function vendor.ScannerItemModel.prototype:IsSelected(row)
	return self.index[row].selected
end

--[[
	Returns the map with the indices of the selected items.
--]]
function vendor.ScannerItemModel.prototype:GetSelectedItems()
	_EnsureSorting(self)
	if (self.updateCount ~= self.selectedUpdateCount) then
    	wipe(self.selectedIndex)
    	for i=1,#self.index do
    		if (self.index[i].selected) then
    			table.insert(self.selectedIndex, i)
    		end
    	end
    	self.selectedUpdateCount = self.updateCount
    end
	return self.selectedIndex
end

--[[
	Registers a function that will be called, if the model has been updated.
--]]
function vendor.ScannerItemModel.prototype:RegisterUpdateListener(func, arg)
	table.insert(self.updateListeners, {func = func, arg = arg})
end

--[[
	Remove all known items.
--]]
function vendor.ScannerItemModel.prototype:Clear()
	wipe(self.index)
end
